Imports System
Imports Microsoft.Win32.SafeHandles
Imports System.Diagnostics
Imports System.Windows.Forms

Namespace WinUsbDemo
	''' <summary>
	'''  Routines for the WinUsb driver.
	'''  </summary>
	'''  <remarks>
	'''  I could find no way to test for a valid winUsbHandle. If the handle is 
	'''  already closed, the application gets an AccessViolationException, which the 
	'''  routines capture and ignore.
	''' </remarks>

	Friend NotInheritable Partial Class WinUsbCommunications
		''' <summary>
		''' Stores information about the WinUSB device.
		''' </summary>

		Private Const ModuleName As String = "WinUSB Device"

		Public Class DeviceInfo
			Friend BulkInPipe As Byte
			Friend BulkOutPipe As Byte
			Friend InterruptInPipe As Byte
			Friend InterruptOutPipe As Byte
			Friend DeviceSpeed As UInt32
		End Class

		'''  <summary>
		'''  Closes the device handle obtained with CreateFile and frees resources.
		'''  </summary>
		'''  <param name="deviceHandle"> Device handle obtained with CreateFile </param>
		'''  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>

		Friend Sub CloseDeviceHandle(ByVal deviceHandle As SafeFileHandle, ByVal winUsbHandle As IntPtr)
			Try
				NativeMethods.WinUsb_Free(winUsbHandle)

				If deviceHandle IsNot Nothing Then
					If Not(deviceHandle.IsInvalid) Then
						deviceHandle.Close()
					End If
				End If
			Catch ex As AccessViolationException
				' The handle might be closed.

			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Sub

		'''  <summary>
		'''  Provides a central mechanism for exception handling.
		'''  Displays a message box that describes the exception.
		'''  </summary>
		'''  
		'''  <param name="Name"> the module where the exception occurred. </param>
		'''  <param name="e"> the exception </param>

		Friend Shared Sub DisplayException(ByVal Name As String, ByVal e As Exception)
			Try
				'  Create an error message.

				Dim message As String = "Exception: " & e.Message & Environment.NewLine & "Module: " & Name & Environment.NewLine & "Method: " & e.TargetSite.Name

				Const caption As String = "Unexpected Exception"

				MessageBox.Show(message, caption, MessageBoxButtons.OK)
				Debug.Write(message)
			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Sub

		'''  <summary>
		'''  Initiates a Control Read transfer. Data stage is device to host.
		'''  </summary>
		''' 
		'''  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		'''  <param name="dataStage"> The received data </param>
		'''  
		'''  <returns>
		'''  True on success, False on failure.
		'''  </returns>

		Friend Function DoControlReadTransfer(ByVal winUsbHandle As IntPtr, ByRef dataStage() As Byte) As Boolean
			Dim bytesReturned As UInt32 = 0

			Try
				'  Vendor-specific request to an interface with device-to-host Data stage.

				Dim setupPacket As NativeMethods.WINUSB_SETUP_PACKET
				setupPacket.RequestType = &HC1

				'  The request number that identifies the specific request.

				setupPacket.Request = 2

				'  Command-specific value to send to the device.
				' For a request directed to an interface, the WinUSB driver uses this field to specify the interface number
				' and will ignore a value set here.
				' For a request directed to the device, you can use this field for vendor-defined purposes.

				setupPacket.Index = 0

				'  Number of bytes in the request's Data stage.

				setupPacket.Length = Convert.ToUInt16(dataStage.Length)

				'  Command-specific value to send to the device.

				setupPacket.Value = 0

				' ***
				'  winusb function 

				'  summary
				'  Initiates a control transfer.

				'  paramaters
				'  Device handle returned by WinUsb_Initialize.
				'  WINUSB_SETUP_PACKET structure 
				'  Buffer to hold the returned Data-stage data.
				'  Number of data bytes to read in the Data stage.
				'  Number of bytes read in the Data stage.
				'  Null pointer for non-overlapped.

				'  returns
				'  True on success.
				'  ***            

				Dim success = NativeMethods.WinUsb_ControlTransfer(winUsbHandle, setupPacket, dataStage, Convert.ToUInt16(dataStage.Length), bytesReturned, IntPtr.Zero)
				Return success

			Catch ex As AccessViolationException
				' The handle might be closed.

				Return False
			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Function

		'''  <summary>
		'''  Initiates a Control Write transfer. Data stage is host to device.
		'''  </summary>
		'''  
		'''  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param> 
		'''  <param name="dataStage"> The data to send. </param>
		'''  
		'''  <returns>
		'''  True on success, False on failure.
		'''  </returns>

		Friend Function DoControlWriteTransfer(ByVal winUsbHandle As IntPtr, ByVal dataStage() As Byte) As Boolean
			Dim bytesReturned As UInteger = 0
			Dim value = Convert.ToUInt16(0)

			Try
				'  Vendor-specific request to an interface with host-to-device Data stage.

				Dim setupPacket As NativeMethods.WINUSB_SETUP_PACKET
				setupPacket.RequestType = &H41

				'  The request number that identifies the specific request.

				setupPacket.Request = 1

				'  Command-specific value to send to the device.
				' For a request directed to an interface, the WinUSB driver uses this field to specify the interface number.
				' and will ignore a value set here.
				' For a request directed to the device, you can use this field for vendor-defined purposes.

				setupPacket.Index = 0

				'  Number of bytes in the request's Data stage.

				setupPacket.Length = Convert.ToUInt16(dataStage.Length)

				'  Command-specific value to send to the device.

				setupPacket.Value = value

				' ***
				'  winusb function 

				'  summary
				'  Initiates a control transfer.

				'  parameters
				'  Device handle returned by WinUsb_Initialize.
				'  WINUSB_SETUP_PACKET structure 
				'  Buffer containing the Data-stage data.
				'  Number of data bytes to send in the Data stage.
				'  Number of bytes sent in the Data stage.
				'  Null pointer for non-overlapped.

				'  Returns
				'  True on success.
				'  ***

				Dim success = NativeMethods.WinUsb_ControlTransfer(winUsbHandle, setupPacket, dataStage, Convert.ToUInt16(dataStage.Length), bytesReturned, IntPtr.Zero)
				Return success
			Catch ex As AccessViolationException
				' The handle might be closed.
				Return False
			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Function

		'''  <summary>
		'''  Requests a handle with CreateFile.
		'''  </summary>
		'''  
		'''  <param name="devicePathName"> Returned by SetupDiGetDeviceInterfaceDetail 
		'''  in an SP_DEVICE_INTERFACE_DETAIL_DATA structure. </param>
		'''  
		'''  <returns>
		'''  The handle.
		'''  </returns>

		Friend Function GetDeviceHandle(ByVal devicePathName As String) As SafeFileHandle
			Try
				' ***
				' API function

				'  summary
				'  Retrieves a handle to a device.

				'  parameters 
				'  Device path name returned by SetupDiGetDeviceInterfaceDetail
				'  Type of access requested (read/write).
				'  FILE_SHARE attributes to allow other processes to access the device while this handle is open.
				'  Security structure. Using Null for this may cause problems under Windows XP.
				'  Creation disposition value. Use OPEN_EXISTING for devices.
				'  Flags and attributes for files. The winsub driver requires FILE_FLAG_OVERLAPPED.
				'  Handle to a template file. Not used.

				'  Returns
				'  A handle or INVALID_HANDLE_VALUE.
				' ***

				Dim deviceHandle As SafeFileHandle = FileIo.NativeMethods.CreateFile(devicePathName, (FileIo.NativeMethods.GENERIC_WRITE Or FileIo.NativeMethods.GENERIC_READ), FileIo.NativeMethods.FILE_SHARE_READ Or FileIo.NativeMethods.FILE_SHARE_WRITE, IntPtr.Zero, FileIo.NativeMethods.OPEN_EXISTING, FileIo.NativeMethods.FILE_ATTRIBUTE_NORMAL Or FileIo.NativeMethods.FILE_FLAG_OVERLAPPED, IntPtr.Zero)
				Return deviceHandle
			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Function

		'''  <summary>
		'''  Initializes a device interface and obtains information about it.
		'''  Calls these winusb API functions:
		'''    WinUsb_Initialize
		'''    WinUsb_QueryInterfaceSettings
		'''    WinUsb_QueryPipe
		'''  </summary>
		'''  
		'''  <param name="deviceHandle"> Device handle obtained with CreateFile </param>
		'''  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		'''  <param name="myDeviceInfo"> devInfo structure for the device </param>	
		'''  <param name="pipeTimeout"> desired timeout in milliseconds for transfers </param>
		'''  
		'''  <returns>
		'''  True on success, False on failure.
		'''  </returns>

		Friend Function InitializeDevice(ByVal deviceHandle As SafeFileHandle, ByRef winUsbHandle As IntPtr, ByRef myDeviceInfo As DeviceInfo, ByVal pipeTimeout As UInt32) As Boolean
			Try
				Dim ifaceDescriptor As NativeMethods.USB_INTERFACE_DESCRIPTOR
				ifaceDescriptor.bLength = 0
				ifaceDescriptor.bDescriptorType = 0
				ifaceDescriptor.bInterfaceNumber = 0
				ifaceDescriptor.bAlternateSetting = 0
				ifaceDescriptor.bNumEndpoints = 0
				ifaceDescriptor.bInterfaceClass = 0
				ifaceDescriptor.bInterfaceSubClass = 0
				ifaceDescriptor.bInterfaceProtocol = 0
				ifaceDescriptor.iInterface = 0

				Dim pipeInfo As NativeMethods.WINUSB_PIPE_INFORMATION
				pipeInfo.PipeType = 0
				pipeInfo.PipeId = 0
				pipeInfo.MaximumPacketSize = 0
				pipeInfo.Interval = 0

				' ***
				'  winusb function 

				'  summary
				'  get a handle for communications with a winusb device        '

				'  parameters
				'  Handle returned by CreateFile.
				'  Device handle to be returned.

				'  returns
				'  True on success.
				'  ***

				Dim success = NativeMethods.WinUsb_Initialize(deviceHandle, winUsbHandle)

				If success Then
					' ***
					'  winusb function 

					'  summary
					'  Get a structure with information about the device interface.

					'  parameters
					'  handle returned by WinUsb_Initialize
					'  alternate interface setting number
					'  USB_INTERFACE_DESCRIPTOR structure to be returned.

					'  returns
					'  True on success.

					success = NativeMethods.WinUsb_QueryInterfaceSettings(winUsbHandle, 0, ifaceDescriptor)

					If success Then
						'  Get the transfer type, endpoint number, and direction for the interface's
						'  bulk and interrupt endpoints. Set pipe policies.

						' ***
						'  winusb function 

						'  summary
						'  returns information about a USB pipe (endpoint address)

						'  parameters
						'  Handle returned by WinUsb_Initialize
						'  Alternate interface setting number
						'  Number of an endpoint address associated with the interface. 
						'  (The values count up from zero and are NOT the same as the endpoint address
						'  in the endpoint descriptor.)
						'  WINUSB_PIPE_INFORMATION structure to be returned

						'  returns
						'  True on success   
						' ***

						For i = 0 To ifaceDescriptor.bNumEndpoints - 1
							NativeMethods.WinUsb_QueryPipe(winUsbHandle, 0, Convert.ToByte(i), pipeInfo)

							If ((pipeInfo.PipeType = NativeMethods.USBD_PIPE_TYPE.UsbdPipeTypeBulk) And UsbEndpointDirectionIn(pipeInfo.PipeId)) Then
								myDeviceInfo.BulkInPipe = pipeInfo.PipeId

								SetPipePolicy(winUsbHandle, myDeviceInfo.BulkInPipe, Convert.ToUInt32(NativeMethods.POLICY_TYPE.IGNORE_SHORT_PACKETS), Convert.ToByte(False))

								SetPipePolicy(winUsbHandle, myDeviceInfo.BulkInPipe, Convert.ToUInt32(NativeMethods.POLICY_TYPE.PIPE_TRANSFER_TIMEOUT), pipeTimeout)
							ElseIf ((pipeInfo.PipeType = NativeMethods.USBD_PIPE_TYPE.UsbdPipeTypeBulk) And UsbEndpointDirectionOut(pipeInfo.PipeId)) Then
								myDeviceInfo.BulkOutPipe = pipeInfo.PipeId

								SetPipePolicy(winUsbHandle, myDeviceInfo.BulkOutPipe, Convert.ToUInt32(NativeMethods.POLICY_TYPE.IGNORE_SHORT_PACKETS), Convert.ToByte(False))

								SetPipePolicy(winUsbHandle, myDeviceInfo.BulkOutPipe, Convert.ToUInt32(NativeMethods.POLICY_TYPE.PIPE_TRANSFER_TIMEOUT), pipeTimeout)
							ElseIf (pipeInfo.PipeType = NativeMethods.USBD_PIPE_TYPE.UsbdPipeTypeInterrupt) And UsbEndpointDirectionIn(pipeInfo.PipeId) Then
								myDeviceInfo.InterruptInPipe = pipeInfo.PipeId

								SetPipePolicy(winUsbHandle, myDeviceInfo.InterruptInPipe, Convert.ToUInt32(NativeMethods.POLICY_TYPE.IGNORE_SHORT_PACKETS), Convert.ToByte(False))

								SetPipePolicy(winUsbHandle, myDeviceInfo.InterruptInPipe, Convert.ToUInt32(NativeMethods.POLICY_TYPE.PIPE_TRANSFER_TIMEOUT), pipeTimeout)
							ElseIf (pipeInfo.PipeType = NativeMethods.USBD_PIPE_TYPE.UsbdPipeTypeInterrupt) And UsbEndpointDirectionOut(pipeInfo.PipeId) Then
								myDeviceInfo.InterruptOutPipe = pipeInfo.PipeId

								SetPipePolicy(winUsbHandle, myDeviceInfo.InterruptOutPipe, Convert.ToUInt32(NativeMethods.POLICY_TYPE.IGNORE_SHORT_PACKETS), Convert.ToByte(False))

								SetPipePolicy(winUsbHandle, myDeviceInfo.InterruptOutPipe, Convert.ToUInt32(NativeMethods.POLICY_TYPE.PIPE_TRANSFER_TIMEOUT), pipeTimeout)
							End If
						Next i
					End If
				End If
				Return success

			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Function

		'''  <summary>
		'''  Is the current operating system Windows XP or later?
		'''  The WinUSB driver requires Windows XP or later.
		'''  </summary>
		''' 
		'''  <returns>
		'''  True if Windows XP or later, False if not.
		'''  </returns>

		Friend Function IsWindowsXpOrLater() As Boolean
			Try
				Dim myEnvironment = Environment.OSVersion

				'  Windows XP is version 5.1.

				Dim versionXp = New Version(5, 1)

				If myEnvironment.Version >= versionXp Then
					Return True
				End If
				Return False

			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Function

		'''  <summary>
		'''  Gets a value that corresponds to a USB_DEVICE_SPEED. 
		'''  </summary>
		'''  
		'''  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		'''  <param name="myDevInfo"> devInfo structure for the device </param>

		Friend Shared Function QueryDeviceSpeed(ByVal winUsbHandle As IntPtr, ByRef myDevInfo As DeviceInfo) As Boolean
			Dim length As UInt32 = 1
			Dim speed = New Byte(0){}

			Try
				' ***
				'  winusb function 

				'  summary
				'  Get the device speed. 
				'  (Normally not required but can be nice to know.)

				'  parameters
				'  Handle returned by WinUsb_Initialize
				'  Requested information type.
				'  Number of bytes to read.
				'  Information to be returned.

				'  returns
				'  True on success.
				' ***           			

				Dim success = NativeMethods.WinUsb_QueryDeviceInformation(winUsbHandle, NativeMethods.DEVICE_SPEED, length, speed(0))

				If success Then
					myDevInfo.DeviceSpeed = Convert.ToUInt32(speed(0))
				End If

				Return success
			Catch ex As AccessViolationException
				' The handle might be closed.
				Return False

			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Function

		'''  <summary>
		'''  Attempts to read data from a bulk IN endpoint.
		'''  </summary>
		'''  
		'''  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		'''  <param name="myDeviceInfo"> devInfo structure for the device </param>		
		'''  <param name="bytesToRead"> Number of bytes to read. </param>
		'''  <param name="dataBuffer"> Buffer for storing the bytes read. </param>
		'''  <param name="bytesRead"> Number of bytes read. </param>
		'''  <param name="success"> Success or failure status. </param>
		'''  
		Friend Sub ReceiveDataViaBulkTransfer(ByVal winUsbHandle As IntPtr, ByVal myDeviceInfo As DeviceInfo, ByVal bytesToRead As UInt32, ByRef dataBuffer() As Byte, ByRef bytesRead As UInt32, ByRef success As Boolean)
			Try
				' ***
				'  winusb function 

				'  summary
				'  Attempts to read data from a device interface.

				'  parameters
				'  Device handle returned by WinUsb_Initialize.
				'  Endpoint address.
				'  Buffer to store the data.
				'  Maximum number of bytes to return.
				'  Number of bytes read.
				'  Null pointer for non-overlapped.

				'  Returns
				'  True on success.
				' ***

					success = NativeMethods.WinUsb_ReadPipe(winUsbHandle, myDeviceInfo.BulkInPipe, dataBuffer, bytesToRead, bytesRead, IntPtr.Zero)
			Catch ex As AccessViolationException
				' The handle might be closed.

			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Sub

		'''  <summary>
		'''  Attempts to read data from an interrupt IN endpoint. 
		'''  </summary>
		'''  
		'''  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		'''  <param name="myDeviceInfo"> devInfo structure for the device </param>		
		'''  <param name="bytesToRead"> Number of bytes to read. </param>
		'''  <param name="dataBuffer"> Buffer for storing the bytes read. </param>
		'''  <param name="bytesRead"> Number of bytes read. </param>
		'''  
		'''  <returns> true on success, false on failure </returns>
		'''  
		Friend Function ReceiveDataViaInterruptTransfer(ByVal winUsbHandle As IntPtr, ByVal myDeviceInfo As DeviceInfo, ByVal bytesToRead As UInt32, ByRef dataBuffer() As Byte, ByRef bytesRead As UInt32) As Boolean
			Try
				' ***
				'  winusb function 

				'  summary
				'  Attempts to read data from a device interface.

				'  parameters
				'  Device handle returned by WinUsb_Initialize.
				'  Endpoint address.
				'  Buffer to store the data.
				'  Maximum number of bytes to return.
				'  Number of bytes read.
				'  Null pointer for non-overlapped.

				'  Returns
				'  True on success.
				' ***

				Dim success As Boolean = NativeMethods.WinUsb_ReadPipe(winUsbHandle, myDeviceInfo.InterruptInPipe, dataBuffer, bytesToRead, bytesRead, IntPtr.Zero)

				Return success
			Catch ex As AccessViolationException
				' The handle might be closed.
				Return False


			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Function

		'''  <summary>
		'''  Attempts to send data via a bulk OUT endpoint.
		'''  </summary>
		'''  
		'''  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		'''  <param name="myDeviceInfo"> devInfo structure for the device </param> 		
		'''  <param name="bytesToWrite"> Number of bytes to write. </param>
		'''  <param name="dataBuffer"> Buffer containing the bytes to write. </param>
		'''  <param name="bytesWritten"> The number of bytes written </param>
		'''  <param name="success"> True on success </param>

		Friend Sub SendDataViaBulkTransfer(ByVal winUsbHandle As IntPtr, ByVal myDeviceInfo As DeviceInfo, ByVal bytesToWrite As UInt32, ByVal dataBuffer() As Byte, ByRef bytesWritten As UInt32, ByRef success As Boolean)
			Try
				' ***
				'  winusb function 

				'  summary
				'  Attempts to write data to a device interface.

				'  parameters
				'  Device handle returned by WinUsb_Initialize.
				'  Endpoint address.
				'  Buffer with data to write.
				'  Number of bytes to write.
				'  Number of bytes written.
				'  IntPtr.Zero for non-overlapped I/O.

				'  Returns
				'  True on success.
				'  ***

					success = NativeMethods.WinUsb_WritePipe(winUsbHandle, myDeviceInfo.BulkOutPipe, dataBuffer, bytesToWrite, bytesWritten, IntPtr.Zero)
			Catch ex As AccessViolationException
				' The handle might be closed.


			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Sub

		'''  <summary>
		'''  Attempts to send data via an interrupt OUT endpoint.
		'''  </summary>
		'''
		'''  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>  
		'''  <param name="myDeviceInfo"> devInfo structure for the device </param> 		
		'''  <param name="bytesToWrite"> Number of bytes to write. </param>
		'''  <param name="dataBuffer"> Buffer containing the bytes to write. </param>
		'''  <param name="bytesWritten"> The number of bytes written </param>		
		'''  
		'''  <returns>
		'''  True on success, False on failure.
		'''  </returns>

		Friend Function SendDataViaInterruptTransfer(ByVal winUsbHandle As IntPtr, ByVal myDeviceInfo As DeviceInfo, ByVal bytesToWrite As UInt32, ByVal dataBuffer() As Byte, ByRef bytesWritten As UInt32) As Boolean
			Try
				' ***
				'  winusb function 

				'  summary
				'  Attempts to write data to a device interface.

				'  parameters
				'  Device handle returned by WinUsb_Initialize.
				'  Endpoint address.
				'  Buffer with data to write.
				'  Number of bytes to write.
				'  Number of bytes written.
				'  IntPtr.Zero for non-overlapped I/O.

				'  Returns
				'  True on success.
				'  ***

				Dim success = NativeMethods.WinUsb_WritePipe(winUsbHandle, myDeviceInfo.InterruptOutPipe, dataBuffer, bytesToWrite, bytesWritten, IntPtr.Zero)

				Return success
			Catch ex As AccessViolationException
				' The handle might be closed.
				Return False


			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Function

		'''  <summary>
		'''  Sets pipe policy.
		'''  Used when the value parameter is a Byte (all except PIPE_TRANSFER_TIMEOUT).
		'''  </summary>
		''' 
		'''  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		''' <param name="pipeId"> Pipe to set a policy for. </param>
		'''  <param name="policyType"> POLICY_TYPE member. </param>
		'''  <param name="value"> Policy value. </param>
		'''  
		'''  <returns>
		'''  True on success, False on failure.
		'''  </returns>
		'''  
		Private Function SetPipePolicy(ByVal winUsbHandle As IntPtr, ByVal pipeId As Byte, ByVal policyType As UInt32, ByVal value As Byte) As Boolean
			Try
				' ***
				'  winusb function 

				'  summary
				'  sets a pipe policy 

				'  parameters
				'  handle returned by WinUsb_Initialize
				'  identifies the pipe
				'  POLICY_TYPE member.
				'  length of value in bytes
				'  value to set for the policy.

				'  returns
				'  True on success 
				' ***

				Dim success = NativeMethods.WinUsb_SetPipePolicy(winUsbHandle, pipeId, policyType, 1, value)

				Return success
			Catch ex As AccessViolationException
				' The handle might be closed.

				Return False


			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Function

		'''  <summary>
		'''  Sets pipe policy.
		'''  Used when the value parameter is a UInt32 (PIPE_TRANSFER_TIMEOUT only).
		'''  </summary>
		''' 
		''' <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		''' <param name="pipeId"> Pipe to set a policy for. </param>
		'''  <param name="policyType"> POLICY_TYPE member. </param>
		'''  <param name="value"> Policy value. </param>
		'''  
		'''  <returns>
		'''  True on success, False on failure.
		'''  </returns>
		'''  
		Private Function SetPipePolicy(ByVal winUsbHandle As IntPtr, ByVal pipeId As Byte, ByVal policyType As UInt32, ByVal value As UInt32) As Boolean
			Dim success = False
			Try
				' ***
				'  winusb function 

				'  summary
				'  sets a pipe policy 

				'  parameters
				'  handle returned by WinUsb_Initialize
				'  identifies the pipe
				'  POLICY_TYPE member.
				'  length of value in bytes
				'  value to set for the policy.

				'  returns
				'  True on success 
				' ***

				success = NativeMethods.WinUsb_SetPipePolicy1(winUsbHandle, pipeId, policyType, 4, value)

				Return success
			Catch ex As AccessViolationException
				' Handle might be closed.

				Return success

			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Function

		'''  <summary>
		'''  Is the endpoint's direction IN (device to host)?
		'''  </summary>
		'''  
		'''  <param name="endpointAddress"> The endpoint address. </param>
		'''  
		'''  <returns>
		'''  True if IN (device to host), False if OUT (host to device)
		'''  </returns> 

		Private Function UsbEndpointDirectionIn(ByVal endpointAddress As Int32) As Boolean
			Try

				Dim directionIn = False

				If ((endpointAddress And &H80) = &H80) Then
					directionIn = True
				End If
				Return directionIn

			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Function

		'''  <summary>
		'''  Is the endpoint's direction OUT (host to device)?
		'''  </summary>
		'''  
		'''  <param name="addr"> The endpoint address. </param>
		'''  
		'''  <returns>
		'''  True if OUT (host to device, False if IN (device to host)
		'''  </returns>

		Private Function UsbEndpointDirectionOut(ByVal addr As Int32) As Boolean
			Try
				Dim directionOut = False

				If ((addr And &H80) = 0) Then
					directionOut = True
				End If
				Return directionOut


			Catch ex As Exception
				DisplayException(ModuleName, ex)
				Throw
			End Try
		End Function

	
	End Class
End Namespace
